#ifndef LEBAI_KDL_H_
#define LEBAI_KDL_H_

#include <stdbool.h>
#include <string.h>

#include <array>

class MovingAverageFilter {
 public:
  MovingAverageFilter() : input_(0.0), out_(0.0), index_(0), size_(100) {}

  void resetSize(unsigned int n) {
    if (!n) n = 1;
    if (n > buffer_.max_size()) n = buffer_.max_size();
    index_ = 0;
    size_ = n;
  }

  void resetData(double data) {
    out_ = data;
    for (unsigned int j = 0; j < size_; ++j) {
      buffer_[j] = data;
    }
  }

  double& push(double new_data) {
    unsigned int current = (index_ + size_) % size_;
    // std::cout << current << " " << new_data(i) << "                       "
    //           << size_[i] << "\n";
    double d = (new_data - buffer_[current]) / static_cast<double>(size_);
    out_ += d;
    buffer_[current] = new_data;
    index_++;
    return out_;
  }

  unsigned int size() { return size_; }

 protected:
  std::array<double, 200> buffer_;
  double input_;
  double out_;
  // KDL::JntArray last_out_;
  unsigned int index_;
  unsigned int size_;
};

struct jointTpPlanner_t {
  double target_vel;
  double final_vel;

  double pos_cmd;  /* position command */
  double max_vel;  /* velocity limit */
  double maxacc;   /* acceleration limit */
  double curr_pos; /* current position */
  double curr_vel; /* current velocity */
  double cycle_time;
  MovingAverageFilter filter;
  double curr_pos_output;
  int enable;
};

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Lebai kdl
 *
 * @author Yonnie Lu
 * @email zhangyong.lu@lebai.ltd
 */

typedef struct kdl_config {
  bool use_friction_compensator;  // = true;
} kdl_config;

typedef struct dh_t {
  double a;
  double alpha;
  double d;
  double theta;
} dh_t;

typedef struct d_param_t {
  double Ixx;    // = 0.0;
  double Iyy;    // = 0.0;
  double Izz;    // = 0.0;
  double Ixy;    // = 0.0;
  double Ixz;    // = 0.0;
  double Iyz;    // = 0.0; /**< I** inertia matrix */
  double mass;   // = 0.0; /**< Mass */
  double com_x;  // = 0.0,
  double com_y;  // = 0.0,
  double com_z;  // = 0.0;                            /**< Center of mass */
} d_param_t;

typedef struct gravity_t {
  double x;
  double y;
  double z;
} gravity_t;

typedef struct poseRPY_t {
  double r;
  double p;
  double y;
} poseRPY_t;

typedef struct posePosition_t {
  double x;
  double y;
  double z;
} posePosition_t;

typedef struct poseFrame_t {
  poseRPY_t rotation;
  posePosition_t pos;

  double vel;
} poseFrame_t;

typedef struct segment_t {
  dh_t dh;
  d_param_t dyn;
} segment_t;

typedef struct joint_data_t {
  double q;
  double v;
  double acc;
  double q_cmd;
  double v_cmd;
  double acc_cmd;
  double torque;
  double extra;  // such as: with_load torque
} joint_data_t;

typedef struct tcp_dynamic_sample_data_t {
  joint_data_t joints[6];  // each joint data
  int joint_index;         // current move joint index
} tcp_dynamic_sample_data_t;

// TODO: update to cog_t
typedef struct cog_t {
  double x;
  double y;
  double z;
} cog_t;

typedef struct payload_t {
  double mass;
  cog_t cog;
} payload_t;

typedef struct KDL_t {
  segment_t segments[16];
  size_t segments_size;
  gravity_t gravity;
  double torque_kp[6];
  double torque_kv[6];
} KDL_t;

typedef struct joint_dynamic {
  double torque;    // algorithm torque
  double intertia;  // joint equivalent intertia
} joint_dynamic;

typedef struct joint_friction_param_t {
  double plus_viscous_friction_k;
  double plus_static_friction;
  double minus_viscous_friction_k;
  double minus_static_friction;
} joint_friction_param_t;

// y = kx + b
typedef struct least_squares_result {
  double k;
  double b;
} least_squares_result;

typedef struct LebaiKDL LebaiKDL;

LebaiKDL* LebaiKDL__new();
LebaiKDL* LebaiKDL__delete(LebaiKDL* self);
// init lebai kdl
int LebaiKDL__init_kdl(LebaiKDL* self, KDL_t* kdl,
                       joint_friction_param_t* j_frictions, kdl_config config);
// clear lebai kdl cache
int LebaiKDL__clear_kdl_cache(LebaiKDL* self);
// update gravity
int LebaiKDL__update_gravity(LebaiKDL* self, gravity_t gravity);
// update tcp segments
int LebaiKDL__update_tcp(LebaiKDL* self, poseFrame_t tcp);
// set payload
int LebaiKDL__set_payload(LebaiKDL* self, payload_t payload);
// get number of joints
int LebaiKDL__get_num_of_joints(LebaiKDL* self);
// get number of segments
int LebaiKDL__get_num_of_segments(LebaiKDL* self);
// fk
int LebaiKDL__forword_kinematics(LebaiKDL* self, poseFrame_t* dest,
                                 double joint[]);
// ik
int LebaiKDL__inverse_kinematics(LebaiKDL* self, poseFrame_t dest,
                                 double* joint, int* iter_count);
// ik for 3pra
int LebaiKDL__inverse_kinematics_3pra(LebaiKDL* self, poseFrame_t dest,
                                      double* joint);
// new kinematic line
int LebaiKDL__path_line(LebaiKDL* self, poseFrame_t start, poseFrame_t end,
                        double eq_radius);
// new kinematic arc
int LebaiKDL__path_arc(LebaiKDL* self, poseFrame_t start, posePosition_t center,
                       poseFrame_t end, double alpha, double eq_radius);
// new kinematic sequence rounded composite
int LebaiKDL__path_rounded_composite(LebaiKDL* self, poseFrame_t route[],
                                     size_t len, double rr_radius,
                                     double eq_radius);
// init path / joint velocity profile
int LebaiKDL__set_vel_profile(LebaiKDL* self, double max_v, double max_acc,
                              double cur_v, double cur_acc, double final_v,
                              double cycle_time);
// Calculate current trapezoiadal pose according to current target_v(LebaiKDL*
// self, which should multiple velocity scale factor) 1: calculate end; 0: not
// end
int LebaiKDL__cal_trap_pose(LebaiKDL* self, poseFrame_t* pose, double targe_v,
                            bool re_calculate);
// Calculate trapezoiadal joint pose
int LebaiKDL__cal_trap_joint_pose(LebaiKDL* self, jointTpPlanner_t* tp);
// Calculate the mass and com of tcp things
payload_t LebaiKDL__cal_tcp_dynamics(LebaiKDL* self,
                                     tcp_dynamic_sample_data_t data[],
                                     size_t len);
// Compensate gravity
int LebaiKDL__gravity_compensate(LebaiKDL* self, const joint_data_t input[],
                                 joint_dynamic output[]);
// Inverse dynamic calculator
int LebaiKDL__inverse_dyn_calculator(LebaiKDL* self, const joint_data_t input[],
                                     joint_dynamic output[]);
// The least-squares solution for x,y
least_squares_result LebaiKDL__least_squares_solution(LebaiKDL* self,
                                                      double* x_data,
                                                      size_t x_size,
                                                      double* y_data,
                                                      size_t y_size);
// Pose frame to homogeneous matrix
void LebaiKDL__frame_2_homo_matrix(LebaiKDL* self, poseFrame_t pose, double* h);
// Calculate the translation(x, y, z) for a transform. result is tcp
int LebaiKDL__cal_tcp_translation(LebaiKDL* self, const poseFrame_t poses[],
                                  size_t size, double tol, poseFrame_t* tcp);

#ifdef __cplusplus
};
#endif

#endif
