#ifndef LEBAI_UTIL_H_
#define LEBAI_UTIL_H_

#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * 机器人模型
 */
typedef struct Robotic Robotic;

/**
 * 直角坐标系（笛卡尔坐标系）
 *
 * $P(x, y, z)$
 */
typedef struct CartesianPosition {
  double x;
  double y;
  double z;
} CartesianPosition;

/**
 * 圆柱坐标系（极坐标系）
 *
 * $P(\rho, \theta, z)$
 */
typedef struct CylindricalPosition {
  /**
   * 径向距离 $\rho$
   */
  double r;
  /**
   * 方位角 $\theta$，也作 $\phi$, $\varphi$
   */
  double theta;
  /**
   * 高度 $z$, 也作 $h$
   */
  double z;
} CylindricalPosition;

/**
 * D-H 参数
 */
typedef struct DHParam {
  double d;
  double a;
  /**
   * $ \alpha $
   */
  double alpha;
  /**
   * $ \theta $
   */
  double theta;
} DHParam;

/**
 * 姿态（欧拉角=ZYX旋转角）
 */
typedef struct RpyPose {
  /**
   * roll, Z
   */
  double r;
  /**
   * pitch, Y
   */
  double p;
  /**
   * yaw, X
   */
  double y;
} RpyPose;

/**
 * 位姿（笛卡尔坐标+欧拉角）
 */
typedef struct XyzRpy {
  CartesianPosition p;
  RpyPose r;
} XyzRpy;

/**
 * 直角坐标系 => 圆柱坐标系
 */
void cart2pol(const CartesianPosition *from, CylindricalPosition *to);

/**
 * 直角坐标系的变化量 => 圆柱坐标系的变化量
 */
void cart2pol_delta(const CartesianPosition *from,
                    const CartesianPosition *delta,
                    CylindricalPosition *to_delta);

/**
 * 直角坐标系的变化量 => 圆柱坐标系
 */
void cart2pol_delta_to(const CartesianPosition *from,
                       const CartesianPosition *delta,
                       CylindricalPosition *to);

/**
 * 圆柱坐标系 => 直角坐标系
 */
void pol2cart(const CylindricalPosition *from, CartesianPosition *to);

/**
 * 圆柱坐标系的变化量 => 直角坐标系的变化量
 */
void pol2cart_delta(const CylindricalPosition *from,
                    const CylindricalPosition *delta,
                    CartesianPosition *to_delta);

/**
 * 圆柱坐标系的变化量 => 直角坐标系
 */
void pol2cart_delta_to(const CylindricalPosition *from,
                       const CylindricalPosition *delta,
                       CartesianPosition *to);

/**
 * 向机器人模型添加一个关节的 D-H 参数
 */
void robotic__add_joint(Robotic *robot, const DHParam *joint);

/**
 * 获取手爪位姿（相对于基座）
 */
XyzRpy robotic__get_claw(Robotic *robot);

/**
 * 获取手爪位姿（相对于 TCP）
 */
XyzRpy robotic__get_claw_to_tcp(Robotic *robot);

/**
 * 获取法兰盘位姿（相对于基座）
 */
XyzRpy robotic__get_flange(Robotic *robot);

/**
 * 获取 TCP 位姿（相对于基座）
 */
XyzRpy robotic__get_tcp(Robotic *robot);

/**
 * 获取 TCP 位姿（相对于法兰盘）
 */
XyzRpy robotic__get_tcp_to_flange(Robotic *robot);

/**
 * 对机器人进行正解
 */
void robotic__kinematics_forward(Robotic *robot);

/**
 * 沿基座方向微调
 */
void robotic__move_by_base(Robotic *robot, const XyzRpy *delta);

/**
 * 沿手爪方向微调
 */
void robotic__move_by_claw(Robotic *robot, const XyzRpy *delta);

/**
 * 沿法兰盘方向微调
 */
void robotic__move_by_flange(Robotic *robot, const XyzRpy *delta);

/**
 * 沿 TCP 方向微调
 */
void robotic__move_by_tcp(Robotic *robot, const XyzRpy *delta);

/**
 * 新建机器人模型
 */
Robotic *robotic__new(void);

/**
 * 设置手爪相对于 TCP 的位姿
 */
void robotic__set_claw_to_tcp(Robotic *robot, const XyzRpy *t);

/**
 * 设置法兰盘相对于基座的位姿
 */
void robotic__set_flange(Robotic *robot, const XyzRpy *t);

/**
 * 设置机器人关节转动角度
 */
void robotic__set_joint_theta(Robotic *robot, uint32_t i, double theta);

/**
 * 设置 TCP 相对于法兰盘的位姿
 */
void robotic__set_tcp_to_flange(Robotic *robot, const XyzRpy *t);

/**
 * 以法兰盘位姿为基准，计算 TCP 和手爪位姿
 *
 * 凡是调用过 `robotic__set_xxx` 后，就应该调用该方法进行更新
 */
void robotic__update_on_flange(Robotic *robot);

#ifdef __cplusplus
};
#endif

#endif
